# Copyright 2018 - M. Pecheux
# [Game Of Life] A simple implementation of Conway's Game of Life
# ------------------------------------------------------------------------------
# patterns.py - Util functions to create well-known patterns of Life (they are
#               meant for Conway's rules: B3/S23)
# ==============================================================================

# UTIL FUNCTIONS
# -----------------

def clear_grid(grid, x, y, w, h):
    for xx in range(-1, w + 2):
        for yy in range(-1, h + 2):
            grid[y + yy, x + xx] = 0

# -----------------

# PATTERN FUNCTIONS
# -----------------
# (patterns are anchored NW)
# Still lifes
def block(grid, x, y):
    grid[y,  x] = 1
    grid[y,  x+1] = 1
    grid[y+1,x] = 1
    grid[y+1,x+1] = 1

def beehive(grid, x, y):
    grid[y,  x+1] = 1
    grid[y,  x+2] = 1
    grid[y+1,x] = 1
    grid[y+1,x+3] = 1
    grid[y+2,x+1] = 1
    grid[y+2,x+2] = 1

def loaf(grid, x, y):
    grid[y,  x+1] = 1
    grid[y,  x+2] = 1
    grid[y+1,x] = 1
    grid[y+1,x+3] = 1
    grid[y+2,x+1] = 1
    grid[y+2,x+3] = 1
    grid[y+3,x+2] = 1

def boat(grid, x, y):
    grid[y,  x] = 1
    grid[y,  x+1] = 1
    grid[y+1,x] = 1
    grid[y+1,x+2] = 1
    grid[y+2,x+1] = 1

def tub(grid, x, y):
    grid[y,  x+1] = 1
    grid[y+1,x] = 1
    grid[y+1,x+2] = 1
    grid[y+2,x+1] = 1

# Oscillators
def blinker(grid, x, y):
    grid[y,  x] = 1
    grid[y+1,x] = 1
    grid[y+2,x] = 1

def toad(grid, x, y):
    grid[y,  x+1] = 1
    grid[y,  x+2] = 1
    grid[y,  x+3] = 1
    grid[y+1,x] = 1
    grid[y+1,x+1] = 1
    grid[y+1,x+2] = 1

def beacon(grid, x, y):
    grid[y,  x] = 1
    grid[y,  x+1] = 1
    grid[y+1,x] = 1
    grid[y+2,x+4] = 1
    grid[y+3,x+3] = 1
    grid[y+3,x+4] = 1

def pulsar(grid, x, y):
    grid[y,   x+2] = 1
    grid[y,   x+3] = 1
    grid[y,   x+4] = 1
    grid[y,   x+8] = 1
    grid[y,   x+9] = 1
    grid[y,   x+10] = 1
    grid[y+2, x] = 1
    grid[y+2, x+5] = 1
    grid[y+2, x+7] = 1
    grid[y+2, x+12] = 1
    grid[y+3, x] = 1
    grid[y+3, x+5] = 1
    grid[y+3, x+7] = 1
    grid[y+3, x+12] = 1
    grid[y+4, x] = 1
    grid[y+4, x+5] = 1
    grid[y+4, x+7] = 1
    grid[y+4, x+12] = 1
    grid[y+5, x+2] = 1
    grid[y+5, x+3] = 1
    grid[y+5, x+4] = 1
    grid[y+5, x+8] = 1
    grid[y+5, x+9] = 1
    grid[y+5, x+10] = 1
    grid[y+7, x+2] = 1
    grid[y+7, x+3] = 1
    grid[y+7, x+4] = 1
    grid[y+7, x+8] = 1
    grid[y+7, x+9] = 1
    grid[y+7, x+10] = 1
    grid[y+8, x] = 1
    grid[y+8, x+5] = 1
    grid[y+8, x+7] = 1
    grid[y+8, x+12] = 1
    grid[y+9, x] = 1
    grid[y+9, x+5] = 1
    grid[y+9, x+7] = 1
    grid[y+9, x+12] = 1
    grid[y+10,x] = 1
    grid[y+10,x+5] = 1
    grid[y+10,x+7] = 1
    grid[y+10,x+12] = 1
    grid[y+12,x+2] = 1
    grid[y+12,x+3] = 1
    grid[y+12,x+4] = 1
    grid[y+12,x+8] = 1
    grid[y+12,x+9] = 1
    grid[y+12,x+10] = 1

def pencadecathlon(grid, x, y):
    grid[y,   x+4] = 1
    grid[y+1, x+4] = 1
    grid[y+2, x+3] = 1
    grid[y+2, x+4] = 1
    grid[y+2, x+5] = 1
    grid[y+5, x+3] = 1
    grid[y+5, x+4] = 1
    grid[y+5, x+5] = 1
    grid[y+6, x+4] = 1
    grid[y+7, x+4] = 1
    grid[y+8, x+4] = 1
    grid[y+9, x+4] = 1
    grid[y+10,x+3] = 1
    grid[y+10,x+4] = 1
    grid[y+10,x+5] = 1
    grid[y+13,x+3] = 1
    grid[y+13,x+4] = 1
    grid[y+13,x+5] = 1
    grid[y+14,x+4] = 1
    grid[y+15,x+4] = 1

# Spaceships
def glider(grid, x, y, direction=None):
    if direction is None:
        raise RuntimeError('Spaceship has incorrect direction.')

    grid[y,  x+1] = 1
    if direction == 'left':
        grid[y+1,x] = 1
    elif direction == 'right':
        grid[y+1,x+2] = 1
    grid[y+2,x] = 1
    grid[y+2,x+1] = 1
    grid[y+2,x+2] = 1

def spaceship(grid, x, y, direction=None):
    if direction is None:
        raise RuntimeError('Spaceship has incorrect direction.')

    if direction == 'left':
        grid[y,  x+1] = 1
        grid[y,  x+4] = 1
        grid[y+1,x] = 1
        grid[y+2,x] = 1
        grid[y+2,x+4] = 1
        grid[y+3,x] = 1
        grid[y+3,x+1] = 1
        grid[y+3,x+2] = 1
        grid[y+3,x+3] = 1
    elif direction == 'right':
        grid[y,  x] = 1
        grid[y,  x+3] = 1
        grid[y+1,x+4] = 1
        grid[y+2,x] = 1
        grid[y+2,x+4] = 1
        grid[y+3,x+1] = 1
        grid[y+3,x+2] = 1
        grid[y+3,x+3] = 1
        grid[y+3,x+4] = 1

# Guns
def gun_gosper_glider(grid, x, y):
    grid[y+4,x] = 1
    grid[y+4,x+1] = 1
    grid[y+5,x] = 1
    grid[y+5,x+1] = 1

    grid[y+2,x+34] = 1
    grid[y+2,x+35] = 1
    grid[y+3,x+34] = 1
    grid[y+3,x+35] = 1

    grid[y,  x+24] = 1

    grid[y+1,x+22] = 1
    grid[y+1,x+24] = 1

    grid[y+2,x+12] = 1
    grid[y+2,x+13] = 1
    grid[y+2,x+20] = 1
    grid[y+2,x+21] = 1

    grid[y+3,x+11] = 1
    grid[y+3,x+15] = 1
    grid[y+3,x+20] = 1
    grid[y+3,x+21] = 1

    grid[y+4,x+10] = 1
    grid[y+4,x+16] = 1
    grid[y+4,x+20] = 1
    grid[y+4,x+21] = 1

    grid[y+5,x+10] = 1
    grid[y+5,x+14] = 1
    grid[y+5,x+16] = 1
    grid[y+5,x+17] = 1
    grid[y+5,x+22] = 1
    grid[y+5,x+24] = 1

    grid[y+6,x+10] = 1
    grid[y+6,x+16] = 1
    grid[y+6,x+24] = 1

    grid[y+7,x+11] = 1
    grid[y+7,x+15] = 1

    grid[y+8,x+12] = 1
    grid[y+8,x+13] = 1

def make_pattern(grid, x, y, type=None):
    if type is None:
        raise RuntimeError('Unknown pattern.')

    if type == 'block':
        clear_grid(grid, x, y, 2, 2)
        block(grid, x, y)
    elif type == 'beehive':
        clear_grid(grid, x, y, 4, 3)
        beehive(grid, x, y)
    elif type == 'loaf':
        clear_grid(grid, x, y, 4, 4)
        loaf(grid, x, y)
    elif type == 'boat':
        clear_grid(grid, x, y, 3, 3)
        boat(grid, x, y)
    elif type == 'tub':
        clear_grid(grid, x, y, 3, 3)
        tub(grid, x, y)
    elif type == 'blinker':
        clear_grid(grid, x, y, 3, 3)
        blinker(grid, x, y)
    elif type == 'toad':
        clear_grid(grid, x, y, 4, 4)
        toad(grid, x, y)
    elif type == 'beacon':
        clear_grid(grid, x, y, 4, 4)
        beacon(grid, x, y)
    elif type == 'pulsar':
        clear_grid(grid, x, y, 13, 13)
        pulsar(grid, x, y)
    elif type == 'pencadecathlon':
        clear_grid(grid, x, y, 9, 16)
        pencadecathlon(grid, x, y)
    elif type == 'glider_l':
        clear_grid(grid, x, y, 3, 3)
        glider(grid, x, y, 'left')
    elif type == 'glider_r':
        clear_grid(grid, x, y, 3, 3)
        glider(grid, x, y, 'right')
    elif type == 'spaceship_l':
        clear_grid(grid, x, y, 5, 4)
        spaceship(grid, x, y, 'left')
    elif type == 'spaceship_r':
        clear_grid(grid, x, y, 5, 4)
        spaceship(grid, x, y, 'right')
    elif type == 'gun_gosper_glider':
        clear_grid(grid, x, y, 36, 9)
        gun_gosper_glider(grid, x, y)

def demo_patterns(grid, x, y):
    # still lifes
    make_pattern(grid, x,    y, 'block')
    make_pattern(grid, x+10,  y, 'beehive')
    make_pattern(grid, x+20, y, 'loaf')
    make_pattern(grid, x+30, y, 'boat')
    make_pattern(grid, x+40, y, 'tub')
    # oscillators
    make_pattern(grid, x,    y+10, 'blinker')
    make_pattern(grid, x+10, y+10, 'toad')
    make_pattern(grid, x+20, y+10, 'beacon')
    # spaceships
    make_pattern(grid, x,    y+20, 'glider_l')
    make_pattern(grid, x+10, y+20, 'glider_r')
    make_pattern(grid, x+30, y+20, 'spaceship_l')
    make_pattern(grid, x+40, y+20, 'spaceship_r')

# -----------------
